import os
import re
import flask
import json
import dash
from enum import IntEnum

from .printf import printf
from .time import time_ms


class DEBUG_LEVEL(IntEnum):
    NONE = 0
    REQUESTS = 1
    VERBOSE = 2

DASH_DEBUG = int(os.environ.get('DASH_DEBUG', 0))

class DebugFormatter:

    def __init__(self, req):
        self.req_inputs = req.get('inputs', [])
        self.req_output = req['output']

        # log.info('req_inputs=%s', json.dumps(self.req_inputs))
        # log.info('req_output=%s', json.dumps(self.req_output))


    def req2str(self):
        """Convert request inputs dictionary into a string for display

        Typical input list:

            [
                {'id': 'test#page1#btn', 'property': 'n_clicks', 'value': 1}
                {'id': 'test#redirect', 'property': 'href', 'value': '/test/route'}
            ]

        Output

            test#page1#btn.n_clicks="1", test#redirect.href="/test/route"

        Arguments:
            req_inputs {list} -- List of input attributes and associated values

        Returns:
            str -- String for display
        """

        input_list = []
        for input in self.req_inputs:
            s = '{}.{}'.format(input['id'], input['property'])
            if 'value' in input:
                s += ':"{}"'.format(input['value'])
            input_list.append(s)

        input_list = ', '.join(input_list)

        outputs = re.sub(r'\.\.\.', ', ', self.req_output)
        outputs = re.sub(r'\.\.', '', outputs)

        return '{} -> [{}]'.format(input_list, outputs)


    def outputs2str(self, resp):

        resp = json.loads(resp.data, encoding="utf-8")['response']

        # log.info('resp=%s', json.dumps(resp))

        outputs = re.sub(r'\.\.\.', ', ', self.req_output)
        outputs = re.sub(r'\.\.', '', outputs)
        outputs = outputs.split(', ')

        element_list = []
        output_index = 0

        for key, output in resp.items():
            attr_list = []

            for attr, value in output.items():
                attr_list.append('"{}"'.format(value))

            attr_values = ', '.join(attr_list)
            element_list.append('{} -> {}'.format(attr_values, outputs[output_index]))
            output_index += 1

        # switch="admin#login" -> spa#router.switch, title="Dash/SPA:Admin login" -> spa#title.title

        element_list = ', '.join(element_list)

        return element_list



class DashDebug(dash.Dash):

    count = 0
    tlast = time_ms()

    def get_index(self):
        count = DashDebug.count + 1
        DashDebug.count = count
        return count

    def get_dt(self):
        tlast = DashDebug.tlast
        DashDebug.tlast = tnow = time_ms()
        return tnow - tlast


    def dispatch(self):
        """_dash-update-component
        
        Returns:
            [type] -- [description]
        """

        body = flask.request.get_json()
        formatter = DebugFormatter(body)

        # Create input report

        count = self.get_index()

        if DASH_DEBUG > DEBUG_LEVEL.NONE:

            input_list = formatter.req2str()

            if self.get_dt() > 1000:
                printf('\n')

            printf('%03d req %s\n', count, input_list)

        else:
            input_length = len(json.dumps(formatter.req_inputs))
            output_length = len(json.dumps(formatter.req_output))
            printf('%03d req %d bytes\n', count, input_length + output_length)


        # Process the inputs

        try:
            resp = super().dispatch()
        except dash.exceptions.PreventUpdate as ex:
            printf('%03d res NOACTION\n', count)
            raise ex

        # Create response report

        if DASH_DEBUG > DEBUG_LEVEL.NONE:

            element_list = formatter.outputs2str(resp)
            printf('%03d res %s\n', count, element_list)

        else:
            printf('%03d res %d bytes\n', count, len(resp.data))


        return resp

    def dependencies(self):
        """_dash-dependencies
        
        Returns:
            [type] -- [description]
        """

        resp = super().dependencies()

        count = self.get_index()
        tmp = json.loads(resp.data, encoding="utf-8")

        if DASH_DEBUG == DEBUG_LEVEL.VERBOSE:

            printf('%03d _dash-dependencies %s\n', count, json.dumps(tmp, indent=2))
        else:
            printf('%03d _dash-dependencies %d bytes\n', count, len(resp.data))

        return resp

    def serve_layout(self):
        """_dash-layout

        Returns:
            [type] -- [description]
        """

        resp = super().serve_layout()

        count = self.get_index()
        tmp = json.loads(resp.data, encoding="utf-8")

        if DASH_DEBUG == DEBUG_LEVEL.VERBOSE:
      
            printf('%03d _dash-layout %s\n', count, json.dumps(tmp, indent=2))
        else:
            printf('%03d _dash-layout %d bytes\n', count, len(resp.data))

        return resp




from utils import log

import dash
from dash.dependencies import DashDependency
import dash_html_components as html
import dash_core_components as dcc
import dash_bootstrap_components as dbc


from .spa_components import SpaComponents

DashDependency.id = property(lambda self: self.component_id)

# Dash supports a ''No Update' option on callback outputs, inject __str__() method
# to get it to display correctly when logging

def __no_update_str__(self):
    return 'NOUPDATE'

setattr(dash.dash._NoUpdate, '__str__', __no_update_str__)
setattr(dash.dash._NoUpdate, '__repr__', __no_update_str__)

class SinglePageApp:

    def __init__(self, _dash, navitems=None, pages=None, title='Dash/SPA'):
        self.dash = _dash
        self.navitems = navitems
        self.title = title
        self.routes = {}
        self.endpoints = {}

    def run_server(self, debug=True, threaded=True):
        self.init()
        self.dash.run_server(debug=debug, threaded=threaded)
        # self.dash.run_server(debug=debug, threaded=threaded, dev_tools_serve_dev_bundles=True)

    def init(self):
        self._layout = self.page_layout()
        self.dash.layout = self.layout
        return self.dash

    def layout(self):
        """Return the dash component tree

        We rebuild the application layout for each refresh. This is
        experimental and may have side effects

        Returns:
            Object -- The root Dash element of a tree of elements
        """

        log.info('SPA: refresh')

        try:

            # Clear the Dash layout & callback state prior rebuilding, it's
            # as if this is the first call prior to calling dash.run_server()


            self.dash.callback_map = {}
            self.dash._layout = None

            # Rebuild everything

            layout = self.page_layout()

        finally:

            # Allow dash to call us again on the next refresh

            self.dash._layout = self.layout

        return layout


    def page_layout(self):
        """layout the application's Dash components

        This method is called by Dash on start-up The applications
        component tree is built and all associated callbacks
        are visited.

        Returns:9
            Object -- The root Dash element of a tree of elements
        """

        spa = SpaComponents('spa')
        _outer = self

        def just_layouts(routes):
            result = {}
            for route, el in routes.items():
                result[route] = el.dash_layout
            return result

        def hash_route(route):
            if route is not None:
                route = route.replace('/', '-')
                return route[1:]
            return None

        # Iterate all registered routes calling the layout
        # function for each route

        routes = {}

        for route, el in self.routes.items():
            try:
                el.dash_layout = el.layout()
                routes[hash_route(route)] = el
            except Exception as ex:
                log.error('Route %s, layout failed: %s', route, ex)
                raise ex

        # Gather the Dash layouts for each of the routes

        layouts = just_layouts(routes)

        # Add an undefined 404 route

        layouts['undefined'] = self.show404()

        (hash_routes, children) = zip(*layouts.items())

        router = spa.LayoutRouter(children, routes=hash_routes, id='router')

        self.page_title = self.title

        title = spa.PageTitle(id='title', title=self.page_title)

        navbar = self.navBar(self.navitems) if self.navitems else None

        @self.dash.callback([router.output.switch, title.output.title], [SpaComponents.url.input.href])
        def _cb_root(href):
            title = spa.NOUPDATE

            url = spa.urlsplit(href)

            pathname = url.path

            log.info('_cb_root[href] %s', pathname)

            hash_pathname = hash_route(pathname)

            if hash_pathname in routes:

                el = routes[hash_pathname]

                if el.login_required:
                    if not self.user_logged_in():
                        login_view = self.login_manager.login_view
                        hash_pathname = self.url_for(login_view)

                if el.anonymouse_only:
                    if self.user_logged_in():
                        hash_pathname = 'undefined'

                if el.validate and not el.validate(href):
                    hash_pathname = 'undefined'


                if el.title:
                    new_title = self.get_title(self.title, el.title)
                    if self.page_title != new_title:
                        title = self.page_title = new_title

            else:
                hash_pathname = 'undefined'

            log.info('router.output.switch: %s title:%s', hash_pathname, title)

            return hash_pathname, title

        # Create the dash layout

        layout = self.page_body(navbar, router)

        layout.children.insert(0, SpaComponents.url)
        layout.children.insert(0, title)

        log.info('SPA: page layout - done')

        return layout

    def get_title(self, title, subtitle):
        return '{}:{}'.format(title, subtitle)

    def page_body(self, navbar, router):
        return html.Div([
            navbar,
            html.Br(),
            html.Div([
                html.Div([
                    html.Div(className="col-md-2"),
                    html.Div(router, className="col-md-8"),
                    html.Div(className="col-md-2")
                ], className='row')
            ], className="container-fluid"),
            html.Div(id='null'),
            self.footer()
        ])

    def footer_text(self):
        return 'Dash/SPA'

    def footer(self):
        return html.Footer([
            html.Div([
                html.P(self.footer_text(), className='text-center font-italic', style={'marginTop': 10})
            ], className='containers')
        ], className='footer')


    def show404(self):
        return html.Div([
            html.Div([
                html.Div([
                    html.Div([
                        html.H1('Oops!'),
                        html.H2('404 Not Found'),
                        html.Div('Sorry, an error has occurred, Requested page not found!', className='error-details'),
                        html.Div([

                            dcc.Link([
                                html.Span(className='fa fa-home'),
                                ' Take Me Home'
                            ], href='/', className='btn btn-secondary btn-lg'),

                            dcc.Link([
                                html.Span(className='fa fa-envelope'),
                                ' Contact Support'
                            ], href='/support', className='btn btn-secondary btn-lg'),



                        ], className='error-actions')
                    ], className='error-template')
                ], className='col-md-12')
            ], className='row')
        ], className='container')

    def navBar(self, navitems, dark=True, color='secondary'):
        """Return the navbar layout, constructed from the supplied dictionary definition

        Returns:
            [object] -- Navbar ready for display
        """

        def item(item):

            def ep_split():
                ep = item["endpoint"]
                ep = ep.split('?')
                return ep[0], ep[1] if len(ep) > 1 else None

            endpoint, querystring = ep_split()
            href = self.url_for(endpoint)

            login_required = item.get('login_required')
            if login_required is not None:
                if self.user_logged_in() != login_required:
                    return None

            if querystring:
                href = href + '?' + querystring

            if 'icon' in item:
                return dbc.NavItem([
                    dbc.NavLink([html.I(className=item['icon']), ' ' + item['title']], href=href)
                ])
            else:
                return dbc.NavItem(dbc.NavLink(item['title'], href=href))

        def get_items(side):
            if navitems and side in navitems:
                items = navitems[side]

                return dbc.Nav([
                    item(x) for x in items
                ])
            return None

        def brand():
            if navitems and 'brand' in navitems:
                brand = navitems['brand']
                return dbc.NavbarBrand(html.Strong(brand['title']), href=brand['href'])
            return None

        navbar = dbc.Navbar(
            children=[
                brand(),
                dbc.NavItem(get_items('left'), className='navbar-nav mr-auto'),
                dbc.NavItem(get_items('right'), className='navbar-nav ml-auto')
            ], className="navbar-default", dark=dark, color=color, expand="md")

        return navbar

    def register_blueprint(self, blueprint, url_prefix=None):
        log.info('register_blueprint %s', url_prefix)

        blueprint.set_pathname(url_prefix)
        blueprint.set_spa_app(self)

        for route, layout_function in blueprint.routes.items():

            full_route = url_prefix + route if url_prefix else route

            if full_route.endswith('/'):
                full_route = full_route[:-1]

            if full_route.startswith('//'):
                full_route = full_route[1:]

            ep = blueprint.name + route.replace('/', '.').rstrip('.')
            log.info('route %s', full_route)
            self.routes[full_route] = layout_function
            self.endpoints[ep] = full_route


    def url_for(self, endpoint):
        try:
            return self.endpoints[endpoint] if endpoint else None
        except Exception:
            raise Exception("Unable to resolve endpoint '{}'".format(endpoint))

    def enable_login_manager(self, login_manager, login_view):

        self.login_manager = login_manager
        self.login_manager.login_view = login_view
        self.login_manager.init_app(self.dash.server)

    def user_logged_in(self):
        """Return True if user is valid and is not anonymouse
        """
        log.info('user_logged_in')
        if self.login_manager:
            try:
                user = self.login_manager.reload_user()
                return user and user.is_anonymous is False
            except Exception:
                pass

        return False



import os
import json


def read_config(config=None):
    if config is None:
        config = os.environ['DASH_SPA_ENV'] if 'DASH_SPA_ENV' in os.environ else 'default'
    file_path = 'config/{}.json'.format(config)
    try:
        with open(file_path) as json_file:
            data = json.load(json_file)
        return data
    except Exception:
        return {}

# https://stackoverflow.com/questions/1305532/convert-nested-python-dict-to-object

class Config(object):
    def __init__(self, d):
        for a, b in d.items():
            if isinstance(b, (list, tuple)):
                setattr(self, a, [Config(x) if isinstance(x, dict) else x for x in b])
            else:
                setattr(self, a, Config(b) if isinstance(b, dict) else b)

def get(conf):
    conf = conf.split('.')
    config = _config
    for c in conf:
        if c in config:
            config = config[c]
        else:
            config = {}
    obj = Config(config)
    return obj

_config = read_config()



"""dash-spa - Dash Single Page Application (SPA) Framework"""

from .spa_dependency import SpaDependency
from .spa_components import SpaComponents
from .spa import SinglePageApp
from .spa_blueprint import Blueprint


__version__ = '0.0.1'
__author__ = 'Steve Jones <jonesst2608@gmail.com>'
__all__ = []



# 1. imports of your dash app
import dash
import dash_html_components as html


# 2. give each testcase a tcid, and pass the fixture
# as a function argument, less boilerplate
def test_bsly001_falsy_child(dash_duo):

    # 3. define your app inside the test function
    app = dash.Dash(__name__)
    app.layout = html.Div(id="nully-wrapper", children=0)

    # 4. host the app locally in a thread, all dash server configs could be
    # passed after the first app argument
    dash_duo.start_server(app)

    # 5. use wait_for_* if your target element is the result of a callback,
    # keep in mind even the initial rendering can trigger callbacks
    dash_duo.wait_for_text_to_equal("#nully-wrapper", "0", timeout=4)

    # 6. use this form if its present is expected at the action point
    assert dash_duo.find_element("#nully-wrapper").text == "0"

    # 7. to make the checkpoint more readable, you can describe the
    # acceptance criterion as an assert message after the comma.
    assert dash_duo.get_logs() == [], "browser console should contain no error"

    # 8. visual testing with percy snapshot
    dash_duo.percy_snapshot("bsly001-layout")



import sys

from .singleton import singleton
from  .logger import logging, log
from .dash_debug import DashDebug
from .printf import printf
from .email_valid import email_valid
from .time import time_ms

from .helpers import scrub_locals

from .config import read_config, get



import logging

from utils import config

log_options = config.get('logging')

logging.basicConfig(
    level=log_options.level,
    # format='%(levelname)s %(asctime)s.%(msecs)03d %(module)10s/%(lineno)-5d %(message)s'
    format='%(levelname)s %(module)10s/%(lineno)-5d %(message)s'
)

log = logging.getLogger()



import dash_html_components as html
from .user import blueprint as spa

from flask_login import current_user


def big_center(text):
    return html.H2(text, className='display-3 text-center')

@spa.route('/profile', title='User')
def profile():

    if current_user and not current_user.is_anonymous:
        name = current_user.name
    else:
        name = 'Guest'

    return html.Header([
        big_center("Dash/SPA Welcomes"),
        big_center(name)
    ], className='jumbotron my-4')



import pystache
import smtplib
from email.mime.text import MIMEText

from utils import config, log

options = config.get('mail_options')
options = options.__getattribute__(options.active)

header_template = """
Subject: {{subject}}
To: {{receiver}}
From: {{sender}}
"""

class TemplateMailer:

    def __init__(self, template, args):
        self.email_msg = pystache.render(template, args)

    @property
    def smt_transport(self):
        if options.secure:
            return smtplib.SMTP_SSL(options.host, options.port)
        else:
            return smtplib.SMTP(options.host, options.port)

    def send(self, sender, receiver, subject):
        try:
            msg = MIMEText(self.email_msg)
            msg['Subject'] = subject
            msg['From'] = sender
            msg['To'] = receiver

            log.info('sending email to %s ...', receiver)
            with self.smt_transport as server:
                server.login(options.auth.user, options.auth.password)
                server.send_message(msg)
                log.info('Done')

        except Exception as ex:
            log.info('email send failed %s', ex)



import re

import dash
from dash.development.base_component import Component
from dash.dependencies import DashDependency

from app import app

# Simple helper to get the Dash components identifier
#
#   my_checkbox.component_id
#
#   can now be just:
#
#   my_checkbox.id

DashDependency.id = property(lambda self: self.component_id)

class DashIOFactory:
    """Dash Dependency Factory

Inject DashIOFactory instances into the Dash component. The
factories will be accessable as follows:

    div = html.Div(id='xxx') 
    children_attr = div.output.children

Will return the same Dash Dependency Output instance as:

    div = html.Div(id='xxx')
    children_attr = DashDependency.Output('xxx', 'children')

    Raises:
        TypeError: raised id the requested attribute not available on the component
        AttributeError: raised if the associated component has no id

    Returns:
        [obj] -- Dash dependency object Output|Input|State
    """

    def __init__(self, component, iofactory):
        assert hasattr(component, 'id'), "The dash component must have an 'id' attribute"
        self.component = component
        self.iofactory = iofactory


    def is_valid(self, attr):
        if attr not in self.available_properties:
            raise TypeError('`' + self._type + '` has no attribute `' + attr + '`')

    def __getattr__(self, name):
        if not name in self.component.available_properties:
            raise AttributeError("'{}' component '{}' has no attribute '{}'".format(self.component._type, self.component.id, name))

        dio = self.iofactory(self.component.id, name)
        self.__setattr__(name, dio)
        return dio


def input(self):
    if not hasattr(self, '_input'):
        self._input = DashIOFactory(self, dash.dependencies.Input)
    return self._input

def output(self):
    if not hasattr(self, '_output'):
        self._output = DashIOFactory(self, dash.dependencies.Output)
    return self._output

def state(self):
    if not hasattr(self, '_state'):
        self._state = DashIOFactory(self, dash.dependencies.State)
    return self._state

# Inject DashIOFactory instances into the Dash component

Component.input = property(input)
Component.output = property(output)
Component.state = property(state)


def strip(label):
    return re.sub(r"\s+", '_', label).strip().lower()

class SpaDependency:

    def __init__(self, prefix):
        self._prefix = strip(prefix)

    def prefix(self, id):
        return '{}-{}'.format(self._prefix, id) if id else None

    def callback(self, output, inputs=[], state=[]):
        # log.info('register cb %s', [inputs, output])
        return app.callback(output, inputs, state)

    def copy_factory(self, src, dest):
        """Copy Dash I/O Factory

        When a Dash component has been wrapped in additional layout to make
        a composite it is necessary to copy the embedded component I/O definition
        to the outermost component. This will then allow the composite component
        to be referenced in Dash callbacks.

        Arguments:
            src {obj} -- The source Dash component
            dest {obj} -- The destination Dash component

        Returns:
            [Obj] -- The destination component
        """

        def _copy():
            dest.input.component = src.input.component
            dest.output.component = src.output.component
            dest.state.component = src.state.component

        if hasattr(dest, 'id'):
            _copy()
        else:
            dest.id = '#container'
            _copy()
            dest.id = None

        return dest




# https://stackoverflow.com/questions/39723398/how-to-inherit-base-class-with-singleton-in-python

class singleton(object):
    instances = {}
    def __new__(cls, clz = None):
        if clz is None:
            # print ("Creating object for", cls)
            if not cls.__name__ in singleton.instances:
                singleton.instances[cls.__name__] = \
                    object.__new__(cls)
            return singleton.instances[cls.__name__]
        # print (cls.__name__, "creating", clz.__name__)
        singleton.instances[clz.__name__] = clz()
        singleton.first = clz
        return type(clz.__name__, (singleton,), dict(clz.__dict__))


import dash_core_components as dcc
import dash_html_components as html
import dash_bootstrap_components as dbc

import pandas as pd

from .demo import blueprint as spa

global_md = """\
### Global Warming
Global Temperature Time Series. Data are included from the GISS 
Surface Temperature (GISTEMP) analysis and the global component 
of Climate at a Glance (GCAG). Two datasets are provided: 

* Global monthly mean
* Annual mean temperature anomalies in degrees Celsius from 1880 to the present

"""

# Taken from Dash example, see:
# https://dash.plot.ly/datatable

df = pd.read_csv('demo/data/solar.csv')

@spa.route('/solar', title='Solar')
def solar():
    return html.Div([
        html.Div([
            html.Div([], className="col-md-2"),
            html.Div([
                html.H2('US Solar Capacity'),
                html.Br(),
                dbc.Table.from_dataframe(df, striped=True, bordered=True, hover=True),
                html.Div(id='output')
            ], className="col-md-8"),
            html.Div([], className="col-md-2")
        ], className='row'),

        dbc.Row([
            dbc.Col([
                dcc.Link("Global Warming", href=spa.url_for('warming'), className="btn btn-primary float-right")
            ], md=12)

        ])

    ], className="container-fluid")



from .views.view_common import blueprint as admin

from .views.login_view import login, logout
from .views.register_view import register
from .views.verify_view import verify
from .views.forgot_view import forgot

from .login_manager import login_manager



import dash_core_components as dcc
import dash_html_components as html
import dash_bootstrap_components as dbc

import pandas as pd

from .demo import blueprint as spa

global_md = """\
### Global Warming
Global Temperature Time Series. Data are included from the GISS 
Surface Temperature (GISTEMP) analysis and the global component 
of Climate at a Glance (GCAG). Two datasets are provided: 

* Global monthly mean
* Annual mean temperature anomalies in degrees Celsius from 1880 to the present

"""

data = pd.read_csv("demo/data/global-warming.csv")

@spa.route('/warming', title='Warming')
def warming():
    return html.Div([
        dbc.Row([
            dbc.Col([
                dcc.Markdown(global_md),
                html.A("View details", href='https://datahub.io/core/global-temp#readme', className="btn btn-secondary"),
            ], md=3),
            dbc.Col([

                # https://dash.plot.ly/dash-core-components/graph

                dcc.Graph(
                    figure={
                        "data": [{
                            "y": data['Mean'].tolist(),
                            "x": data['Year'].tolist()
                        }],
                        'layout': {
                            'title': 'Global Temperature Change (&#176;C)',
                            'xaxis': {'title': 'Year'}

                        }
                    },
                    config={'displayModeBar': False},

                ),
            ], md=9),

        ]),
        dbc.Row([
            dbc.Col([
                dcc.Link("State Solar", href=spa.url_for('solar'), className="btn btn-primary float-right")
            ], md=12)

        ])
    ])



from dash_spa import spa, Blueprint
import dash_html_components as html
from app import app as dash_app

demo = Blueprint('demo')

NAV_BAR_ITEMS = {
    'brand' : {'title' : 'SPA/Example', 'href' : '/'},
    'left' : [
        {'title' : 'Page1', 'endpoint' : 'demo.page1'},
        {'title' : 'Page2', 'endpoint' : 'demo.page2'},
    ],
}

def big_center(text):
    return html.H2(text, className='display-3 text-center')

@demo.route('/page1')
def test1():
    return html.Div([
        big_center('Multi-page Example'),
        big_center('+'),
        big_center('Page 1'),
    ])

@demo.route('/page2')
def test2():
    return html.Div([
        big_center('Multi-page Example'),
        big_center('+'),
        big_center('Page 2'),
    ])

class CustomSpaApp(spa.SinglePageApp):
    """Example of customising Navbar & Footer

    https://en.wikipedia.org/wiki/Web_colors
    """

    def navBar(self, navitems):
        return super().navBar(navitems, color='DarkSlateGray')

    def footer_text(self):
        return 'Multi-page Example'


def create_app():
    app = CustomSpaApp(dash_app, navitems=NAV_BAR_ITEMS)
    app.register_blueprint(demo, url_prefix='/demo')
    return app

#
# python -m examples.multipage
#
# http://localhost:8050/demo/page1
#

if __name__ == '__main__':
    print('\nvisit: http://localhost:8050/demo/page1\n')
    app = create_app()
    app.run_server(debug=False, threaded=False)



from examples.redirect import create_app

def test_redirect(dash_duo):

    app = create_app().init()
    dash_duo.start_server(app)
    dash_duo.server_url = dash_duo.server_url + '/test/page1'

    # https://dash.plot.ly/testing
    # https://www.testingexcellence.com/css-selectors-selenium-webdriver/
    # https://www.w3schools.com/cssref/css_selectors.asp

    # dash_duo.wait_for_element("style='display:block;'", timeout=30)
    dash_duo.wait_for_style_to_equal("#spa-router-test-page1", "display", "block", timeout=30)

    webelement = dash_duo.find_element("#test-page1-btn")
    assert webelement.is_displayed()



from utils import log
from urllib import parse
from .spa_components import SpaComponents

class Blueprint(SpaComponents):

    @property
    def name(self):
        return self.io._prefix

    def __init__(self, name):
        super().__init__(name)
        self.spa_app = None
        self.routes = {}

    def route(self, rule, **options):

        class _Route:
            def __init__(self, layout):
                self.__dict__ = options
                self.layout = layout

            def __getattr__(self, name):
                if name in self.__dict__:
                    return self[name]
                else:
                    return None


        def decorator(f):
            log.info('add rule %s%s', self.name, rule)
            self.routes[rule] = _Route(f)
            return f

        return decorator

    def set_spa_app(self, spa_app):
        self.spa_app = spa_app

    def url_for(self, endpoint, args=None):
        """Convert endpoint to url

        Arguments:
            endpoint {str} -- Endpoint to convert

        Keyword Arguments:
            args {dict} -- Arguments to be encoded as a querystring (default: {None})

        Returns:
            str -- The converted endpoint
        """

        def _get_ep():
            nonlocal endpoint
            try:
                ep = self.spa_app.url_for(endpoint)
                return ep
            except Exception:
                pass

            if endpoint and not '.' in endpoint:
                endpoint = '{}.{}'.format(self.name, endpoint)

            return self.spa_app.url_for(endpoint)

        ep = _get_ep()
        if args:
            ep += '?' + parse.urlencode(args)

        return ep



from utils import logging, log
import dash_html_components as html

from dash_spa import spa, Blueprint
from app import app as dash_app

test = Blueprint('test')

@test.route('/page1')
def route1():
    spa = test.get_spa('page1')

    btn = spa.Button('Page #2', id='btn')
    redirect = spa.Redirect(id='redirect')

    @spa.callback(redirect.output.href, [btn.input.n_clicks])
    def _cb(clicks):
        red = test.NOUPDATE
        log.info('btn clicks=%s', clicks)
        if clicks:
            red = test.url_for('page2')
        return red

    return html.Div([
        html.H2('Page #1'),
        btn,
        redirect
    ])


@test.route('/page2')
def route2():
    spa = test.get_spa('route2')


    btn = spa.Button('Page #1', id='btn')
    redirect = spa.Redirect(id='redirect')

    @spa.callback(redirect.output.href, [btn.input.n_clicks])
    def _cb(clicks):
        red = test.NOUPDATE
        log.info('btn clicks=%s', clicks)
        if clicks:
            red = test.url_for('page1')
        return red

    return html.Div([
        html.H2('Page #2'),
        btn,
        redirect
    ])

def create_app():
    aps_log = logging.getLogger('werkzeug')
    aps_log.setLevel(logging.ERROR)

    app = spa.SinglePageApp(dash_app)
    app.register_blueprint(test, url_prefix='/test')

    return app

#
# python -m examples.redirect
#
# http://localhost:8050/test/page1
#

if __name__ == '__main__':
    print('\nvisit: http://localhost:8050/test/page1\n')
    app = create_app()
    app.run_server(debug=False, threaded=False)



from utils import config, log
from random import randrange
from .template_mailer import TemplateMailer

from flask_login import LoginManager, UserMixin, login_user, logout_user, current_user
from werkzeug.security import generate_password_hash, check_password_hash
from flask_sqlalchemy import SQLAlchemy

from app import app

VERIFICATION_TEMPLATE = """
Hi {{name}}

You recently requested to create a Dash/SPA account. In order to complete
the registration process please enter the following code:

        {{code}}
                          
This code is only valid for the next 30  minutes. 

If you did not request an account, please ignore this
email or reply to let us know.
"""

FORGOT_TEMPLATE = """

You recently requested to reset the password on your Dash/SPA account. In order
to complete the reset process please enter the following code:

        {{code}}
                          
This code is only valid for the next 30  minutes. 

If you did not request an account, please ignore this
email or reply to let us know.
"""

mail_options = config.get('mail_options')
user_db = config.get('user_db')

def randomCode(length=4):
    return ''.join([chr(randrange(10) + 65) for n in range(length)])

class AdminLoginManager(LoginManager):

    def __init__(self, app=None, add_context_processor=True):
        super().__init__(app, add_context_processor)

        self.code = None
        self.email = None

        app.config['SQLALCHEMY_DATABASE_URI'] = user_db.database_uri
        app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

        self.db = SQLAlchemy()
        self.db.init_app(app)

        self.User = self.user_model(self.db)
        self.user_loader(self.load_user)


    def register(self, name, email, password, terms):

        log.info('register [name: %s, email: %s, password: %s, terms: %s]', name, email, password, terms)

        user = self.User.query.filter_by(email=email).first() # if this returns a user, then the email already exists in database

        if user:
            return False

        self.code = randomCode()
        self.name = name
        self.email = email
        self.password = password

        mailer = TemplateMailer(VERIFICATION_TEMPLATE, {'name' : name, 'code': self.code})
        mailer.send(mail_options.sender, email, 'Password verification')
        return True

    def validate(self, code):

        if self.code and (self.code == code):

            new_user = self.User(email=self.email, name=self.name, password=generate_password_hash(self.password, method='sha256'))

            self.db.session.add(new_user)
            self.db.session.commit()

            return True

        return False

    def login(self, email, password, remember):
        log.info('login [email: %s, password: %s, remember: %s]', email, password, remember)

        user = self.User.query.filter_by(email=email).first()

        if not user or not check_password_hash(user.password, password):
            return False

        log.info('Valid user %s', user)
        login_user(user, remember=remember)
        return True

    def change_password(self, email, password):
        user = self.User.query.filter_by(email=email).first()

        if user:
            user.password = generate_password_hash(password, method='sha256')
            self.db.session.add(user)
            self.db.session.commit()
            return True

        return False

    def forgot(self, email):

        log.info('forgot [email: %s]', email)

        user = self.User.query.filter_by(email=email).first() # if this returns a user, then the email already exists in database

        if user is None:
            return False

        self.code = randomCode(4)
        self.email = email

        log.info('code: %s', self.code)

        mailer = TemplateMailer(FORGOT_TEMPLATE, {'code': self.code})
        mailer.send(mail_options.sender, email, 'Password verification')

        return True

    def forgot_code_valid(self, code, email=None):
        if email and email != self.email:
            return False
        return self.code == code.upper()

    def get_email(self):
        return self.email

    def reload_user(self, user=None):
        log.info('reload_user user=%s', user)
        super().reload_user(user)
        return current_user


    def load_user(self, user_id):
        log.info('load_user: %s', user_id)
        return self.User.query.get(int(user_id))

    def logout_user(self):
        log.info('logout_user')
        logout_user()


    def user_model(self, db):

        class _User(UserMixin, db.Model):
            id = db.Column(db.Integer, primary_key=True) # primary keys are required by SQLAlchemy
            email = db.Column(db.String(100), unique=True)
            password = db.Column(db.String(100))
            name = db.Column(db.String(1000))  

        return _User

login_manager = AdminLoginManager(app.server)



import io
import os
import re

from setuptools import find_packages, setup

HERE = os.path.dirname(os.path.abspath(__file__))

def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding='utf-8') as fd:
        return re.sub(text_type(r':[a-z]+:`~?(.*?)`'), text_type(r'``\1``'), fd.read())


def _get_version():
    """ Get version by parsing _version programmatically """
    packages = find_packages()
    version_ns = {}
    with open(
            os.path.join(HERE, packages[0], "_version.py")
    ) as f:
        exec(f.read(), {}, version_ns)
    version = version_ns["__version__"]
    return version


setup(
    name="dash-spa",
    version=_get_version(),
    description="Dash Single Page Application (SPA) Framework",
    packages=find_packages(),
    long_description=read("landing-page.md"),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        'License :: OSI Approved :: MIT License',
        "Operating System :: OS Independent",
    ],
    url="https://github.com/stevej2608/dash-spa",
    author="Steve Jones",
    author_email="jonesst2608@gmail.com",
    license='MIT',
    extras_require={
        "dev": [
            "pytest >= 3.6",
            "invoke",
            "twine",
            "termcolor"
        ],
    },
)



import colorlover as cl

import dash_core_components as dcc
import dash_html_components as html

import pandas as pd

from dash_spa import SpaComponents

from utils import log

from .demo import blueprint as demo

# Taken from Dash example, see:
# https://github.com/plotly/dash-stock-tickers-demo-app

current_tickers = []

colorscale = cl.scales['9']['qual']['Paired']

try:
    df = pd.read_csv('https://raw.githubusercontent.com/plotly/datasets/master/dash-stock-ticker-demo.csv')
except Exception:
    print("Unable to read 'dash-stock-ticker-demo.csv' from github, no internet connection?")
    exit(0)


@demo.route('/ticker', title='Ticker')
def ticker():
    spa = demo.get_spa('ticker')

    # http://localhost:8050/finance_explorer?tickers=TSLA

    url_args_name = 'tickers'

    stock_ticker_dropdown = spa.Dropdown(
        id='stock_ticker',
        name=url_args_name,
        options=[{'label': s[0], 'value': str(s[1])}
                 for s in zip(df.Stock.unique(), df.Stock.unique())],
        multi=True,
        querystring=True
    )

    graphs = spa.Div(id='graphs')

    # Callback to create the charts for requested tickers

    @spa.callback(graphs.output.children, [stock_ticker_dropdown.input.value])
    def _update_graph(tickers):
        global current_tickers
        log.info('_update_graph, stock_ticker_dropdown.input: %s', tickers)
        if tickers is None:
            tickers = current_tickers
        current_tickers = tickers
        return update_graph(tickers)


    # Callback to update the browser search-bar with the
    # selected tickers

    location = spa.Redirect(id='redirect', refresh=False)

    @spa.callback(location.output.href, [stock_ticker_dropdown.input.value])
    def _update_url(tickers):
        log.info('_update_url, stock_ticker_dropdown.input: %s', tickers)
        href = SpaComponents.NOUPDATE
        if tickers is not None:
            href = demo.url_for('ticker')
            urlargs = '+'.join(tickers)
            search = '?{}={}'.format(url_args_name, urlargs)
            href += search
        return href


    # Layout the page

    return html.Div([
        location,
        html.H2('Finance Explorer'),
        html.Br(),
        stock_ticker_dropdown,
        html.Br(),
        graphs
    ], className="container")

def bbands(price, window_size=10, num_of_std=5):
    rolling_mean = price.rolling(window=window_size).mean()
    rolling_std = price.rolling(window=window_size).std()
    upper_band = rolling_mean + (rolling_std*num_of_std)
    lower_band = rolling_mean - (rolling_std*num_of_std)
    return rolling_mean, upper_band, lower_band

def update_graph(tickers):
    graphs = []

    if not tickers:
        graphs.append(html.H3(
            "Select a stock ticker.",
            style={'marginTop': 20, 'marginBottom': 20}
        ))
    else:
        for i, ticker in enumerate(tickers):

            dff = df[df['Stock'] == ticker]

            candlestick = {
                'x': dff['Date'],
                'open': dff['Open'],
                'high': dff['High'],
                'low': dff['Low'],
                'close': dff['Close'],
                'type': 'candlestick',
                'name': ticker,
                'legendgroup': ticker,
                'increasing': {'line': {'color': colorscale[0]}},
                'decreasing': {'line': {'color': colorscale[1]}}
            }
            bb_bands = bbands(dff.Close)
            bollinger_traces = [{
                'x': dff['Date'], 'y': y,
                'type': 'scatter', 'mode': 'lines',
                'line': {'width': 1, 'color': colorscale[(i*2) % len(colorscale)]},
                'hoverinfo': 'none',
                'legendgroup': ticker,
                'showlegend': True if i == 0 else False,
                'name': '{} - bollinger bands'.format(ticker)
            } for i, y in enumerate(bb_bands)]
            graphs.append(dcc.Graph(
                id=ticker,
                figure={
                    'data': [candlestick] + bollinger_traces,
                    'layout': {
                        'margin': {'b': 0, 'r': 10, 'l': 60, 't': 0},
                        'legend': {'x': 0}
                    }
                }
            ))

    return graphs



from functools import wraps

from urllib import parse
import dash
import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html

import dash_holoniq_components as dhc

from utils import log

from .spa_dependency import SpaDependency


class SPAComponent:

    @property
    def layout(self):
        return self._layout


class SpaComponents:

    url = None

    NOUPDATE = dash.no_update


    def __init__(self, prefix, parent=None):
        self.io = SpaDependency(prefix)
        self.url_pathname = prefix.split('-')[-1]
        self.parent = parent

        if SpaComponents.url is None:
            SpaComponents.url = self.Location()

    def get_spa(self, prefix):
        """Return a new SPAComponent that is a child of this instance

        Arguments:
            prefix {str} -- the postfix to append to the current prefix
        """
        return SpaComponents('{}-{}'.format(self.io._prefix, prefix), self)


    def prefix(self, id):
        return self.io.prefix(id)

    def get_pathname(self):
        pathname = (self.parent.get_pathname() + '/' if self.parent else '') + self.url_pathname
        if not pathname.startswith('/'):
            pathname = '/' + pathname
        return  pathname.rstrip('/')

    def set_pathname(self, pathname):
        self.url_pathname = pathname

    def callback(self, output, inputs=[], state=[]):
        return self.io.callback(output, inputs, state)

    def urlsplit(self, href):

        class URL:
            def __init__(self):
                self.url = parse.urlsplit(href)

                query = parse.urlsplit(href).query
                self.qs = parse.parse_qs(query)

                path = self.url.path.decode("utf-8") if isinstance(self.url.path, bytes) else self.url.path

                if href:
                    tmp = path[1:] if path.startswith('/') else path
                    self.paths = tmp.split('/')
                else:
                    self.paths = []

                self.path = path


            @property
            def path_last(self):
                return self.paths[-1]

        return URL()


    def querystring_args(self, href):
        """URL parser

        The parser returns `None` if the last element of the URL path does not
        match the spa prefix that was passed when the SpaDependency instance
        was created. Assuming the prefix matches then the URL query string is
        returned as a dictionary

        Arguments:
            href {str} -- The URL to be parsed

        Returns:
            [dict] -- K,V pairs of the query string
        """
        results = None
        if href:
            url = self.urlsplit(href)
            if url.path == self.get_pathname():
                results = url.qs
        return results


    def Form(self, children, id=None, preventDefault=True, **kwargs):
        id = self.prefix(id)
        form = dhc.Form(children, id=id, preventDefault=preventDefault, **kwargs)
        return form


    def Div(self, children=None, **kwargs):
        id = kwargs.get('id', None)
        kwargs['id'] = self.prefix(id)
        return html.Div(children, **kwargs)

    def Null(self):
        id = self.prefix('null')
        return html.Div(id=id)


    def Button(self, label=None, id=None, type='button', className="btn btn-primary btn-block", **kwargs):
        """Button"""
        id = self.prefix(id)
        return html.Button(label, id=id, type=type, className=className, **kwargs)


    def ButtonLink(self, label=None, id=None, **kwargs):
        """ButtonLink"""
        io = self.io
        id = io.prefix(id) if id else None

        if href is None:
            href = '#'

        return dhc.ButtonLink(id=id, **kwargs)


    def Checkbox(self, children=None, id=None, checked=False, **kwargs):
        """Checkbox"""
        io = self.io
        id = self.prefix(id)

        checkbox = dbc.Checkbox(id=id, className="form-check-input", checked=checked, **kwargs)

        @io.callback(checkbox.output.key, [checkbox.input.checked])
        def _location_cb(checked):
            self.value = checked
            return SpaComponents.NOUPDATE

        fields = [
            checkbox,
            dbc.Label(children, html_for=id, className="form-check-label")
        ]

        _layout = dbc.FormGroup(fields, check=True)
        io.copy_factory(checkbox, _layout)
        return _layout


    def Location(self, id=None, refresh=False, **kwargs):
        """Location"""

        id = self.prefix(id) if id else 'spa#root#url'
        return dcc.Location(id, refresh=refresh, **kwargs)


    def Dropdown(self, id=None, name=None, querystring=False, **kwargs):
        """Dropdown"""

        io = self.io
        id = io.prefix(id)

        dropdown = dcc.Dropdown(id=id, **kwargs)

        if querystring:

            @io.callback(dropdown.output.value, [SpaComponents.url.input.href])
            def _location_cb(href):
                value = SpaComponents.NOUPDATE

                log.info('dropdown %s: href=%s (%s)', dropdown.id, href, self.get_pathname())

                qs = self.querystring_args(href)

                if qs is not None:
                    if name in qs:
                        value = self.value = qs[name][0].split()

                return value

        return dropdown


    def Input(self, label=None, id=None, type='text', prompt=None, name=None, feedback=None, autoComplete=None, querystring=False, **kwargs):
        """Input"""

        io = self.io
        id = io.prefix(id)

        def add_feedback():
            if feedback:
                return [dbc.FormFeedback(feedback, valid=False)]
            else:
                return []

        ac = 'on' if autoComplete else None

        _value = None

        input = dbc.Input(id=id, name=name, type=type, autoComplete=ac, value=_value, **kwargs)

        if querystring:

            @io.callback(input.output.value, [SpaComponents.url.input.href])
            def _location_cb(href):
                nonlocal _value
                log.info('input %s: href=%s (%s)', input.id, href, self.get_pathname())
                qs = self.querystring_args(href)
                if qs and name in qs:
                    _value = qs[name][0]

                log.info('input %s: value %s', input.id, self.value)

                return _value

        fields = [
            input,
        ]

        if label:
            fields.insert(0, dbc.Label(label))

        if prompt:
            fields.append(dbc.FormText(prompt))

        fields += add_feedback()

        _layout = dbc.FormGroup(fields)

        if id is not None:
            io.copy_factory(input, _layout)

        return _layout


    def PasswordInput(self, label, id=None, prompt=None, feedback=None, autoComplete=None, **kwargs):
        """PasswordWithShow"""

        io = self.io
        id = io.prefix(id)

        def add_feedback():
            if feedback:
                return [dbc.FormFeedback(feedback, valid=False)]
            else:
                return []

        ac = 'on' if autoComplete else None

        input = dhc.PasswordWithShow(id=id, **kwargs)

        fields = [input]

        if label:
            fields.insert(0, dbc.Label(label))

        if prompt:
            fields.append(dbc.FormText(prompt))

        fields += add_feedback()

        _layout = dbc.FormGroup(fields)

        if id is not None:
            io.copy_factory(input, _layout)

        return _layout


    def LayoutRouter(self, children=[], switch=None, routes=[], id=None):
        """LayoutRouter"""
        io = self.io
        return dhc.LayoutRouter(children, id=io.prefix(id), switch=switch, routes=routes)


    def ChildRouter(self, id=None):
        """Encapsulation of the standard Dash mechanism for including dynamic
        layout. New layout is sent to the component's `children` attribute.

        The mechanism has a serious shortcomming in that and callbacks associated
        with the dynamic content will not executed. LayoutRouter may be a better
        alternative.

        Keyword Arguments:
            id {string} -- the components ID (default: {None})

        Returns:
            {obj} -- Component instance with callback I/O helpers
        """
        io = self.io
        return html.Div([], id=io.prefix(id))

    def Redirect(self, id=None, refresh=None, href=None):
        """Redirect"""

        io = self.io
        return dhc.Redirect(id=io.prefix(id), refresh=refresh, href=href)


    def Flash(self, message=None, id=None, className="alert alert-danger", **kwargs):
        """Flash"""
        io = self.io
        return dhc.Alert(message, id=io.prefix(id), className=className, role="alert", **kwargs)


    def PageTitle(self, title=None, id=None):
        """PageTitle"""
        io = self.io
        return dhc.PageTitle(id=io.prefix(id), title=title)
